/***********************************************************************************
 *   Copyright (C) 2009 Valerio Orlandini <valerio.orlandini@gmx.com>
 *   Part of this code is based on the simple sine generator developed by
 *   Nedko Arnaudov <nedko@arnaudov.name> and Andreas Kusterer <malaclypse23@gmx.de>
 *   and on the noise generator by Nick Lamb <njl195@zepler.org.uk> and
 *   James W. Morris <james@jwm-art.net>.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *   Questo programma è software libero, soggetto ai termini della licenza
 *   GNU General Public License. Può essere modificato e ridistribuito
 *   secondo i termini specificati.
 *   Il presente programma è distribuito senza garanzia alcuna.
 *********************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>

#include "include/lv2.h"
#include "include/lv2_event.h"
#include "include/lv2_event_helpers.h"
#include "include/lv2_uri_map.h"
#include "sndgen.c"

#define MIDI_INPUT      0
#define CTRL_1			1
#define CTRL_2			2
#define AUDIO_OUTPUT_FL	3
#define AUDIO_OUTPUT_FR	4
#define PANNER			5

#define tme time(NULL) 

static LV2_Event_Buffer *EventBuffer = NULL;
static LV2_Event_Feature *event_ref = NULL;
extern float sndgen(float a, float b, float c, float d);

struct auds {
	/* this is the series of parameters for the audio output; each audio channel has its own values */
	float * out_fl;
	float * out_fr;
	float * ctrl_1;
	float * ctrl_2;
	float * panner;
	float ramp_fl;
	float ramp_fr;
	float note_volume_fl;
	float note_volume_fr;
	unsigned char note_fl;
	unsigned char note_fr;
	float midi_note_fl[128];
	float midi_note_fr[128];
	int midi_event_id_fl;
	int midi_event_id_fr;
};

static LV2_Handle instantiate (
					const LV2_Descriptor * descriptor,
					double sample_rate,
					const char * bundle_path,
					const LV2_Feature * const * host_features
								)
{
	struct auds * outsig;

	LV2_URI_Map_Feature *map_feature;
	register int i;

	/* generation of a random base note, < 40, > 1040 */
	srand(tme);
	register int rdm = 40 + (rand() / (RAND_MAX/1000));

	/* let's add some other randomness, this time creating a weird scale */
	register double rdm2;
	int exp;
	rdm2 = frexp(rand(), &exp);

	/* memory allocation of the whole structure */
	outsig = malloc(sizeof(struct auds));

	/* setting the initial values */
	outsig->ramp_fl = 0.0;
	outsig->note_fl = 0;
	outsig->note_volume_fl = 0.0;

	outsig->ramp_fr = 0.0;
	outsig->note_fr = 0;
	outsig->note_volume_fr = 0.0;

	/* defining the whole set of notes (69 is the C of the 3rd octave according to the MIDI standard) */
	for (i = 0 ; i < 128 ; i++)
    {
		outsig->midi_note_fl[i] = ((rdm+15) * pow((rdm2+0.9), ((float)i - 69.0) / exp)) / sample_rate;
		outsig->midi_note_fr[i] = ((rdm+10) * pow((rdm2+1.1), ((float)i - 69.0) / exp)) / sample_rate;
    }

	/* the following part is about the interaction between plugin and the host */
	for (i = 0; host_features[i]; ++i)
    {
		if (!strcmp(host_features[i]->URI, "http://lv2plug.in/ns/ext/uri-map"))
	{
		map_feature = host_features[i]->data;
		outsig->midi_event_id_fl = map_feature->uri_to_id(map_feature->callback_data, "http://lv2plug.in/ns/ext/event", "http://lv2plug.in/ns/ext/midi#MidiEvent");
		outsig->midi_event_id_fr = map_feature->uri_to_id(map_feature->callback_data, "http://lv2plug.in/ns/ext/event", "http://lv2plug.in/ns/ext/midi#MidiEvent");
	}
		else if (!strcmp(host_features[i]->URI, "http://lv2plug.in/ns/ext/event"))
    {
		event_ref = host_features[i]->data;
    }
  }

	if (outsig->midi_event_id_fl == 0 || outsig->midi_event_id_fr == 0 || event_ref == NULL)
  {
		printf("Cannot initialize - Impossibile inizializzare/n");
		return NULL;
  }

	return (LV2_Handle)outsig;
}

static inline uint16_t handle_event(
									LV2_Event* ev,
									struct auds * outsig
									)
{
	uint16_t size = 0;

	if (ev->type == 0)
    {
		event_ref->lv2_event_unref(event_ref->callback_data, ev);
    }
	else if (ev->type == outsig->midi_event_id_fl || ev->type == outsig->midi_event_id_fr)
    {
		size = ev->size;
    }

	return size;
}

static inline void generate(
							struct auds * outsig,
							uint32_t start,
							uint32_t end
							)
{
	float * const output_l = outsig->out_fl;
	float * const output_r = outsig->out_fr;
			
	/* panner formulas, in order to have left only for panner -100, and right only for panner +100 */
	float pan_fl = (*outsig->panner / -200) + 0.5;
	float pan_fr = (*outsig->panner / 200) + 0.5;
	  
	/* taking the various wave controls and normalizing them according to optimal ranges */
	float ctrl1 = *(outsig->ctrl_1) / 25;
	float ctrl2 = *(outsig->ctrl_2) / 50;
		
	uint32_t pos;
	
	for (pos = start ; pos < end ; pos++)
    {	  
		/* here we process each sample of the output sound; the controls change the sound properties at the change of their value */
		outsig->ramp_fl += outsig->midi_note_fl[outsig->note_fl];
		outsig->ramp_fl = (outsig->ramp_fl > 1.0) ? outsig->ramp_fl - 2.0 : outsig->ramp_fl;
		output_l[pos] = pan_fl * sndgen(outsig->note_volume_fl, ctrl1, outsig->ramp_fl, ctrl2);

		outsig->ramp_fr += outsig->midi_note_fr[outsig->note_fr];
		outsig->ramp_fr = (outsig->ramp_fr > 1.0) ? outsig->ramp_fr - 2.0 : outsig->ramp_fr;
		output_r[pos] =  pan_fr * sndgen(outsig->note_volume_fr, ctrl1, outsig->ramp_fr, ctrl2);
    }
}

#define outsig ((struct auds *)lv2instance)


static void connect_port(
						LV2_Handle lv2instance,
						uint32_t port,
						void * data
						)
{
	switch (port)
	{
		/* data to be sent for each port specified in the ttl file */
		case MIDI_INPUT:
			EventBuffer = data;
			break;
		case CTRL_1:
			outsig->ctrl_1 = data;
			break;
		case CTRL_2:
			outsig->ctrl_2 = data;
			break;
		case AUDIO_OUTPUT_FL:
			outsig->out_fl = data;
			break;
		case AUDIO_OUTPUT_FR:
			outsig->out_fr = data;
			break;
		case PANNER:
			outsig->panner = data;
			break;
  }
}

static void cleanup(
					LV2_Handle lv2instance
					)
{
	free(outsig);
}

static void run(
					LV2_Handle lv2instance,
					uint32_t sample_count
				)
{
	uint32_t offset = 0;
	LV2_Event_Iterator iterator;
	uint32_t frames_done = 0;
	uint16_t size = 0;
	uint8_t *midi_data;

	lv2_event_begin(&iterator, EventBuffer);

	while (frames_done < sample_count)
	{
		uint32_t to;

		if (lv2_event_is_valid(&iterator))
		{
			size = handle_event(lv2_event_get(&iterator, NULL), outsig);
			to = lv2_event_get(&iterator, NULL)->frames;
			midi_data = (uint8_t *) lv2_event_get(&iterator, NULL) + 12;
		}

		/* note on */
		if (size == 3 && ((*midi_data & 0xF0) == 0x90) && (*(midi_data+1) < 128))
		{
			generate(outsig, offset, to);
			outsig->note_fl = *(midi_data+1);
			outsig->note_volume_fl = 1.0;
			outsig->note_fr = *(midi_data+1);
			outsig->note_volume_fr = 1.0;
			offset = to;
			frames_done += offset;
		}

		/* note off */
		else if (size == 3 && ((*midi_data & 0xF0) == 0x80))
		{
			generate(outsig, offset, to);
			outsig->note_volume_fl = 0.0;
			outsig->note_volume_fr = 0.0;
			offset = to;
			frames_done += offset;
		}

		else
		{
			generate(outsig, offset, sample_count);
			frames_done = sample_count;
		}

		if (lv2_event_is_valid(&iterator))
			{
			lv2_event_increment(&iterator);
			size = 0;
			}
	}
}


static LV2_Descriptor g_lv2descriptor =
{
	/* this is a mandatory part of the LV2 specification, pretty self-explanatory */
	.URI = "http://interferon.drivehq.com/plugins/crazynth_st",
	.instantiate = instantiate,
	.connect_port = connect_port,
	.run = run,
	.cleanup = cleanup,
};

const LV2_Descriptor * lv2_descriptor(
					uint32_t index
						)
{
	if (index == 0)
    {
		return &g_lv2descriptor;
	}

	return NULL;
}
